# coding: utf-8
# pylint: disable=C0301,E1101

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""

DEPRECATED. Newer, leaner workflow in module gbi.

--------

GBI publisher.

Workflow sketch:

Two kinds:

* Fulltext
* References

Two shipment types:

* Dump
* Update

Fulltext is called FZS.

TODO:

For fulltext, we take a Sigel-Kind-Package matrix and information about which
database belong to which package and calculate coverage from there.

For references, we reject all by default and only allow those items, which are
covered by a licence. Might change in the future.

Assets:

* gbi_package_db_map_fulltext.tsv
* gbi_package_db_map_refs.tsv
* gbi_package_sigel_map.json

Assumptions:

* A database is either reference or fulltext related, never both.

A few examples:

Database, that are existent in updates, but not dump:

    $ comm -23 <(taskcat GBIUpdateDatabaseList|sort) <(taskcat GBIDumpDatabaseList|cut -f3|sort)
    CHAN
    CIOD
    EXFO
    FLWA
    FSA
    PCW
    TEC
    WEN

Configuration:

[gbi]

scp-src = username@ftp.example.de:/home/gbi
rsync-options = -e XXX -avzP

"""

import collections
import datetime
import json
import os
import re
import tempfile

import luigi

from gluish.format import TSV
from gluish.intervals import weekly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.benchmark import timed
from siskin.common import Executable
from siskin.sources.amsl import AMSLFilterConfig
from siskin.task import DefaultTask
from siskin.utils import SetEncoder, iterfiles

# UPDATE and DUMP are addition configuration that might move into siskin.ini,
# once it is stable

UPDATES = {
    # Patterns for files containing updates. Change processing code too, if you
    # change this pattern, since it will depend on the capture group.
    "patterns": {
        "fulltext": re.compile(r'kons_sachs_fzs_([0-9]{14,14})_.*.zip'),
        "references": re.compile(r'kons_sachs_(?!fzs)[^_]+_([0-9]{14,14})_.*.zip'),
    }
}

# These files must be nested zips. If this changes, change the dependent tasks
# business logic as well.
DUMP = {
    "files": {
        "fulltext": [
            'mirror/gbi/FZS_NOV_2015.zip'
        ],
        "references": [
            'mirror/gbi/PSYN_NOV_2015.zip',
            'mirror/gbi/SOWI_NOV_2015.zip',
            'mirror/gbi/TECHN_NOV_2015.zip',
            'mirror/gbi/WIWI_NOV_2015.zip',
        ]
    },
    "date": datetime.date(2015, 11, 1),
}


class GBITask(DefaultTask):
    """
    GBI task.
    """
    TAG = 'gbi-deprecated'

    # DUMPTAG: the most recent update date as string.
    DUMPTAG = DUMP['date'].strftime("%Y%m%d%H%M%S")

    def closest(self):
        """ Update weekly. """
        return weekly(self.date)


class GBIDropbox(GBITask):
    """
    Pull down GBI dropbox content.
    """
    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return Executable('rsync', message='https://rsync.samba.org/')

    def run(self):
        target = os.path.join(self.taskdir(), 'mirror')
        shellout("mkdir -p {target} && rsync {rsync_options} {src} {target}",
                 rsync_options=self.config.get(
                     'gbi', 'rsync-options', '-avzP'),
                 src=self.config.get('gbi', 'scp-src'), target=target)

        if not os.path.exists(self.taskdir()):
            os.makedirs(self.taskdir())

        with self.output().open('w') as output:
            for path in iterfiles(target):
                output.write_tsv(path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='filelist'), format=TSV)


class GBIDump(GBITask):
    """
    List of GBI dump nested zips relative to the GBIDropbox taskdir.

    Note: Different for fulltext or references.
    """
    kind = luigi.Parameter(
        default='fulltext', description='fulltext or references')

    def requires(self):
        return GBIDropbox()

    def run(self):
        dropbox = GBIDropbox()
        with self.output().open('w') as output:
            if self.kind not in DUMP.get('files'):
                raise RuntimeError('no such fileset: %s, available: %s' % (
                    self.kind, DUMP.get('files')))
            files = DUMP['files'][self.kind]
            for path in files:
                abspath = os.path.join(dropbox.taskdir(), path)
                output.write_tsv(abspath)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='filelist'), format=TSV)


class GBIDumpDatabaseList(GBITask):
    """
    A list of all database names in data dump, as seen they occur in the zipfiles.

    Note: Different for fulltext or references.
    """
    kind = luigi.Parameter(default='fulltext')

    def requires(self):
        return GBIDump(kind=self.kind)

    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('path',)):
                shellout(r"""7z l {input} |
                             awk '{{print $6}}' |
                             grep "zip$" |
                             cut -d '.' -f1 |
                             awk '{{ print "{kind}\t{input}\t"$0 }}' >> {output}""", kind=self.kind, input=row.path, output=stopover)
        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class GBIDumpXML(GBITask):
    """
    Extract a single Database from a dump file.

    Note: DB can be fulltext or references.
    """
    issue = luigi.Parameter(
        default=GBITask.DUMPTAG, description='tag to use as artificial "Dateissue" for dump')
    db = luigi.Parameter(description='name of the database to extract')

    def requires(self):
        return {'dblist-fulltext': GBIDumpDatabaseList(kind='fulltext'),
                'dblist-references': GBIDumpDatabaseList(kind='references'),
                '7z': Executable(name='7z', message='http://www.7-zip.org/')}

    def run(self):
        archive = None

        with self.input().get('dblist-fulltext').open() as handle:
            for row in handle.iter_tsv(cols=('kind', 'path', 'db')):
                if row.db == self.db:
                    archive = row.path

        with self.input().get('dblist-references').open() as handle:
            for row in handle.iter_tsv(cols=('kind', 'path', 'db')):
                if row.db == self.db:
                    archive = row.path

        if archive is None:
            # an non-existent database name amount to an empty dump file
            self.logger.debug('no such db: %s' % self.db)
            with self.output().open('w') as output:
                pass
            return

        dbzip = shellout(
            "7z x -so {archive} {db}.zip 2> /dev/null > {output}", archive=archive, db=self.db)
        output = shellout("""unzip -p {dbzip} \*.xml 2> /dev/null |
                             iconv -f iso-8859-1 -t utf-8 |
                             LC_ALL=C grep -v "^<\!DOCTYPE GENIOS PUBLIC" |
                             LC_ALL=C sed -e 's@<?xml version="1.0" encoding="ISO-8859-1" ?>@@g' |
                             LC_ALL=C sed -e 's@</Document>@<x-origin>{origin}</x-origin><x-issue>{issue}</x-issue></Document>@' |
                             pigz -c >> {output} """, dbzip=dbzip, origin=archive, issue=self.issue)

        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='xml.gz'))


class GBIDumpIntermediateSchema(GBITask):
    """
    Convert dump XML for DB into IS.
    """
    issue = luigi.Parameter(
        default=GBITask.DUMPTAG, description='tag to use as artificial "Dateissue" for dump')
    db = luigi.Parameter(description='name of the database to extract')

    def requires(self):
        return {
            'file': GBIDumpXML(issue=self.issue, db=self.db),
            'span-import': Executable(name='span-import'),
            'pigz': Executable(name='pigz'),
        }

    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        shellout("span-import -i genios <(unpigz -c {input}) | pigz -c >> {output}",
                 input=self.input().get('file').path, output=stopover)
        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'))


class GBIUpdateList(GBITask):
    """
    All GBI files that contain updates since a given date.

    Note: Different for fulltext or references.
    """
    since = luigi.DateParameter(
        default=DUMP['date'], description='used in filename comparison')
    date = ClosestDateParameter(default=datetime.date.today())
    kind = luigi.Parameter(default='fulltext')

    def requires(self):
        return GBIDropbox(date=self.date)

    def run(self):
        since = self.since.strftime('%Y%m%d%H%M%S')

        with self.input().open() as handle:
            with self.output().open('w') as output:
                for row in sorted(handle.iter_tsv(cols=('path',))):
                    match = UPDATES['patterns'][self.kind].search(row.path)
                    if not match:
                        continue
                    filedate = match.group(1)
                    if filedate < since:
                        continue
                    output.write_tsv(row.path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='filelist'), format=TSV)


class GBIUpdateDatabaseList(GBITask):
    """
    A list of databases contained in the updates. New databases might be added
    and only be visible in the updates.

    Note: Different for fulltext or references.
    """
    since = luigi.DateParameter(
        default=DUMP['date'], description='used in filename comparison')
    date = ClosestDateParameter(default=datetime.date.today())
    kind = luigi.Parameter(default='fulltext')

    def requires(self):
        return GBIUpdateList(date=self.date, since=self.since, kind=self.kind)

    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('path',)):
                shellout(""" unzip -p {input} | grep -o 'DB="[^"]*' | sed -e 's/DB="//g' >> {output} """,
                         input=row.path, output=stopover)
        output = shellout(
            "sort -u {input} > {output} && rm {input}", input=stopover)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class GBIDatabaseList(GBITask):
    """
    All database names, that occur, either in dump, update.

    Note: Different for fulltext or references.
    """
    since = luigi.DateParameter(
        default=DUMP['date'], description='used in filename comparison')
    date = ClosestDateParameter(default=datetime.date.today())
    kind = luigi.Parameter(default='fulltext')

    def requires(self):
        return {'dump': GBIDumpDatabaseList(kind=self.kind),
                'update': GBIUpdateDatabaseList(date=self.date, since=self.since, kind=self.kind)}

    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        output = shellout("cat <(cut -f3 {dump}) {update} | sort -u > {output}",
                          dump=self.input().get('dump').path, update=self.input().get('update').path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class GBIDatabaseType(GBITask):
    """
    Database-Type mapping generated directly from raw data.

    Contains info about all databases, from dumps, updates, references and fulltexts.
    """
    since = luigi.DateParameter(
        default=DUMP['date'], description='used in filename comparison')
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'fulltext': GBIDatabaseList(date=self.date, since=self.since, kind='fulltext'),
                'references': GBIDatabaseList(date=self.date, since=self.since, kind='references')}

    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        shellout(r"""cat {input} | awk '{{ print $0"\treferences"}}' >> {output}""",
                 input=self.input().get('references').path, output=stopover)
        shellout(r"""cat {input} | awk '{{ print $0"\tfulltext"}}' >> {output}""",
                 input=self.input().get('fulltext').path, output=stopover)
        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class GBIDatabaseMap(GBITask):
    """
    Create a map to use with span-import (assets/genios/dbmap.json).
    """
    since = luigi.DateParameter(
        default=DUMP['date'], description='used in filename comparison')
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return GBIDatabaseType(date=self.date, since=self.since)

    def run(self):
        dbmap = collections.defaultdict(set)

        with luigi.LocalTarget(self.assets('gbi_package_db_map_fulltext.tsv'), format=TSV).open() as handle:
            for row in handle.iter_tsv(cols=('package', 'db')):
                dbmap[row.db].add(row.package)
                # Full text means Fachzeitschriften (FZS). FZS is a package as
                # well as an indicator to treat these records according to
                # license agreement.
                dbmap[row.db].add('Fachzeitschriften')

        with luigi.LocalTarget(self.assets('gbi_package_db_map_refs.tsv'), format=TSV).open() as handle:
            for row in handle.iter_tsv(cols=('package', 'db')):
                # Whereas reference databases are check along with a generic
                # holding file later.
                dbmap[row.db].add(row.package)

        with self.output().open('w') as output:
            output.write(json.dumps(dbmap, cls=SetEncoder) + "\n")

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class GBIUpdate(GBITask):
    """
    Concatenate updates.
    """
    since = luigi.DateParameter(
        default=DUMP['date'], description='used in filename comparison')
    date = ClosestDateParameter(default=datetime.date.today())
    kind = luigi.Parameter(default='fulltext')

    def requires(self):
        return GBIUpdateList(date=self.date, kind=self.kind)

    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        with self.input().open() as handle:
            for row in sorted(handle.iter_tsv(cols=('path',))):
                match = UPDATES['patterns'][self.kind].search(row.path)
                if not match:
                    continue
                filedate = match.group(1)
                shellout(""" unzip -p {input} | iconv -f iso-8859-1 -t utf-8 |
                             LC_ALL=C grep -v "^<\!DOCTYPE GENIOS PUBLIC" |
                             LC_ALL=C sed -e 's@<?xml version="1.0" encoding="ISO-8859-1" ?>@@g' |
                             LC_ALL=C sed -e 's@</Document>@<x-origin>{origin}</x-origin><x-group>TBA</x-group><x-issue>{issue}</x-issue></Document>@' >> {output}""",
                         input=row.path, origin=row.path, issue=filedate, output=stopover)

        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='xml'))


class GBIUpdateIntermediateSchema(GBITask):
    """
    Convert the combined updates to ischema.
    """
    since = luigi.DateParameter(
        default=DUMP['date'], description='used in filename comparison')
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return [GBIUpdate(since=self.since, date=self.date, kind='fulltext'),
                GBIUpdate(since=self.since, date=self.date, kind='references')]

    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        for target in self.input():
            shellout(
                """span-import -i genios {input} | pigz -c >> {output}""", input=target.path, output=stopover)
        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'))


class GBIUpdateIntermediateSchemaFiltered(GBITask):
    """
    Filter out items for a given database.
    """
    since = luigi.DateParameter(
        default=DUMP['date'], description='used in filename comparison')
    date = ClosestDateParameter(default=datetime.date.today())
    db = luigi.Parameter(description='name of the database to extract')

    def requires(self):
        return GBIUpdateIntermediateSchema(since=self.since, date=self.date)

    def run(self):
        output = shellout("""jq -r -c '. | select(.["x.package"] == "{db}")' <(unpigz -c {input}) | pigz -c > {output}""",
                          db=self.db, input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'))


class GBIDatabase(GBITask):
    """
    Combine dump and updates for a given database.
    """
    issue = luigi.Parameter(
        default=GBITask.DUMPTAG, description='tag to use as artificial "Dateissue" for dump')
    since = luigi.DateParameter(
        default=DUMP['date'], description='used in filename comparison')
    date = ClosestDateParameter(default=datetime.date.today())
    db = luigi.Parameter(description='name of the database to extract')

    def requires(self):
        return {
            'dump': GBIDumpIntermediateSchema(issue=self.issue, db=self.db),
            'update': GBIUpdateIntermediateSchemaFiltered(since=self.since, date=self.date, db=self.db)
        }

    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        for source, target in self.input().iteritems():
            shellout("cat {input} >> {output}",
                     input=target.path, output=stopover)

        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'))


class GBIIntermediateSchemaByKind(GBITask):
    """
    Combine all references or fulltext databases into a single intermediate schema document.
    """
    issue = luigi.Parameter(
        default=GBITask.DUMPTAG, description='tag to use as artificial "Dateissue" for dump')
    since = luigi.DateParameter(
        default=DUMP['date'], description='used in filename comparison')
    date = ClosestDateParameter(default=datetime.date.today())
    kind = luigi.Parameter(default='fulltext')

    def requires(self):
        prereq = GBIDatabaseList(kind=self.kind)
        luigi.build([prereq])
        with prereq.output().open() as handle:
            for row in handle.iter_tsv(cols=('db',)):
                yield GBIDatabase(db=row.db, issue=self.issue, since=self.since, date=self.date)

    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        for target in self.input():
            shellout("cat {input} >> {output}",
                     input=target.path, output=stopover)
        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'))


class GBIExportByKind(GBITask):
    """
    A SOLR-importable version of GBI, by kind.
    """
    issue = luigi.Parameter(
        default=GBITask.DUMPTAG, description='tag to use as artificial "Dateissue" for dump')
    since = luigi.DateParameter(
        default=DUMP['date'], description='used in filename comparison')
    date = ClosestDateParameter(default=datetime.date.today())
    kind = luigi.Parameter(default='fulltext')

    def requires(self):
        return {
            'file': GBIIntermediateSchemaByKind(issue=self.issue, since=self.since, date=self.date, kind=self.kind),
            'config': AMSLFilterConfig(date=self.date)
        }

    @timed
    def run(self):
        output = shellout("span-tag -c {config} <(unpigz -c {input}) | span-export | pigz -c > {output}",
                          input=self.input().get('file').path, config=self.input().get('config').path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'))


class GBIDatabaseTable(GBITask):
    """
    For a given database name, extract the ID and date.
    """
    issue = luigi.Parameter(
        default=GBITask.DUMPTAG, description='tag to use as artificial "Dateissue" for dump')
    since = luigi.DateParameter(
        default=DUMP['date'], description='used in filename comparison')
    date = ClosestDateParameter(default=datetime.date.today())
    db = luigi.Parameter(description='name of the database to extract')

    def requires(self):
        return GBIDatabase(issue=self.issue, since=self.since, date=self.date, db=self.db)

    def run(self):
        output = shellout(
            r"""jq -r '[.["finc.record_id"], .["x.indicator"]] | @csv' <(unpigz -c {input}) | tr -d '"' | tr ',' '\t' > {output}""", input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class GBIISSNDatabase(GBITask):
    """
    Record ISSN per record per database.
    """
    issue = luigi.Parameter(
        default=GBITask.DUMPTAG, description='tag to use as artificial "Dateissue" for dump')
    since = luigi.DateParameter(
        default=DUMP['date'], description='used in filename comparison')
    date = ClosestDateParameter(default=datetime.date.today())
    db = luigi.Parameter(description='name of the database to extract')

    def requires(self):
        return GBIDatabase(issue=self.issue, since=self.since, date=self.date, db=self.db)

    def run(self):
        output = shellout(r"""jq -r '[.["finc.record_id"], .["rft.issn"][]? // "NOT_AVAILABLE"] | @csv' <(unpigz -c {input}) |
                              tr -d '"' | tr ',' '\t' | awk '{{ print "{db}\t"$0 }}'> {output}""", input=self.input().path, db=self.db)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class GBIPublicationTitleDatabase(GBITask):
    """
    Extract publication title per database.
    """
    issue = luigi.Parameter(
        default=GBITask.DUMPTAG, description='tag to use as artificial "Dateissue" for dump')
    since = luigi.DateParameter(
        default=DUMP['date'], description='used in filename comparison')
    date = ClosestDateParameter(default=datetime.date.today())
    db = luigi.Parameter(description='name of the database to extract')

    def requires(self):
        return GBIDatabase(issue=self.issue, since=self.since, date=self.date, db=self.db)

    def run(self):
        output = shellout(r"""jq -r '[.["finc.record_id"], .["rft.jtitle"]? // "NOT_AVAILABLE"] | @csv' <(unpigz -c {input}) |
                              tr -d '"' | tr ',' '\t' | awk '{{ print "{db}\t"$0 }}' > {output}""", input=self.input().path, db=self.db)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class GBIPublicationTitleWrapper(GBITask, luigi.WrapperTask):
    """
    Just a wrapper to prepare publication title lists.
    """
    issue = luigi.Parameter(
        default=GBITask.DUMPTAG, description='tag to use as artificial "Dateissue" for dump')
    since = luigi.DateParameter(
        default=DUMP['date'], description='used in filename comparison')
    date = ClosestDateParameter(default=datetime.date.today())
    kind = luigi.Parameter(default='fulltext')

    def requires(self):
        prerequisite = GBIDatabaseList(kind=self.kind)
        luigi.build([prerequisite])
        with prerequisite.output().open() as handle:
            for row in handle.iter_tsv(cols=('db',)):
                yield GBIPublicationTitleDatabase(issue=self.issue, since=self.since, date=self.date, db=row.db)

    def output(self):
        return self.input()


class GBIPublicationTitleOverview(GBITask):
    """
    Just combine all publication titles.
    """
    issue = luigi.Parameter(
        default=GBITask.DUMPTAG, description='tag to use as artificial "Dateissue" for dump')
    since = luigi.DateParameter(
        default=DUMP['date'], description='used in filename comparison')
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'references': GBIPublicationTitleWrapper(issue=self.issue, since=self.since, date=self.date, kind='references'),
                'fulltext': GBIPublicationTitleWrapper(issue=self.issue, since=self.since, date=self.date, kind='fulltext')}

    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        for kind, targets in self.input().iteritems():
            for target in targets:
                shellout(""" cat {input} | awk '{{ print "{kind}\t"$0 }}' >> {output} """,
                         input=target.path, output=stopover, kind=kind)
        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class GBITitleDatabase(GBITask):
    """
    Extract title per database, also ISSN.
    """
    issue = luigi.Parameter(
        default=GBITask.DUMPTAG, description='tag to use as artificial "Dateissue" for dump')
    since = luigi.DateParameter(
        default=DUMP['date'], description='used in filename comparison')
    date = ClosestDateParameter(default=datetime.date.today())
    db = luigi.Parameter(description='name of the database to extract')

    def requires(self):
        return GBIDatabase(issue=self.issue, since=self.since, date=self.date, db=self.db)

    def run(self):
        output = shellout(r"""jq -r '[.["finc.record_id"], .["rft.issn"][]? // "NOT_AVAILABLE", .["rft.atitle"]? // "NOT_AVAILABLE"] | @csv' <(unpigz -c {input}) |
                              tr -d '"' | tr ',' '\t' | awk '{{ print "{db}\t"$0 }}' > {output}""", input=self.input().path, db=self.db)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class GBITitleWrapper(GBITask, luigi.WrapperTask):
    """
    Just a wrapper to prepare title lists.

    Note: fulltext or reference.
    """
    issue = luigi.Parameter(
        default=GBITask.DUMPTAG, description='tag to use as artificial "Dateissue" for dump')
    since = luigi.DateParameter(
        default=DUMP['date'], description='used in filename comparison')
    date = ClosestDateParameter(default=datetime.date.today())
    kind = luigi.Parameter(default='fulltext')

    def requires(self):
        prerequisite = GBIDatabaseList(kind=self.kind)
        luigi.build([prerequisite])
        with prerequisite.output().open() as handle:
            for row in handle.iter_tsv(cols=('db',)):
                yield GBITitleDatabase(issue=self.issue, since=self.since, date=self.date, db=row.db)

    def output(self):
        return self.input()


class GBITitleList(GBITask):
    """
    Just a wrapper to prepare title lists.

    Note: fulltext or reference.
    """
    issue = luigi.Parameter(
        default=GBITask.DUMPTAG, description='tag to use as artificial "Dateissue" for dump')
    since = luigi.DateParameter(
        default=DUMP['date'], description='used in filename comparison')
    date = ClosestDateParameter(default=datetime.date.today())
    kind = luigi.Parameter(default='fulltext')

    def requires(self):
        return GBITitleWrapper(issue=self.issue, since=self.since, date=self.date, kind=self.kind)

    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        for target in self.input():
            shellout("cat {input} >> {output}",
                     input=target.path, output=stopover)
        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class GBIDatabaseISSNWrapper(GBITask, luigi.WrapperTask):
    """
    Just a wrapper to prepare ISSN lists.
    """
    issue = luigi.Parameter(
        default=GBITask.DUMPTAG, description='tag to use as artificial "Dateissue" for dump')
    since = luigi.DateParameter(
        default=DUMP['date'], description='used in filename comparison')
    date = ClosestDateParameter(default=datetime.date.today())
    kind = luigi.Parameter(default='fulltext')

    def requires(self):
        prerequisite = GBIDatabaseList(kind=self.kind)
        luigi.build([prerequisite])
        with prerequisite.output().open() as handle:
            for row in handle.iter_tsv(cols=('db',)):
                yield GBIISSNDatabase(issue=self.issue, since=self.since, date=self.date, db=row.db)

    def output(self):
        return self.input()


class GBIISSNList(GBITask):
    """
    A list of ISSN in GBI. About 400.
    """
    issue = luigi.Parameter(
        default=GBITask.DUMPTAG, description='tag to use as artificial "Dateissue" for dump')
    since = luigi.DateParameter(
        default=DUMP['date'], description='used in filename comparison')
    date = ClosestDateParameter(default=datetime.date.today())
    kind = luigi.Parameter(default='fulltext')

    def requires(self):
        return GBIDatabaseISSNWrapper(issue=self.issue, since=self.since, date=self.date, kind=self.kind)

    def run(self):
        issns = set()
        for target in self.input():
            with target.open() as handle:
                for line in handle:
                    parts = line.strip().split('\t')
                    if len(parts) < 3:
                        continue
                    match = re.search(
                        '.*([0-9]{4}-[0-9]{3}[0-9X]).*', parts[2])
                    if match:
                        issns.add(match.group(1))

        with self.output().open('w') as output:
            for issn in sorted(issns):
                output.write_tsv(issn)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class GBIDatabaseWrapper(GBITask, luigi.WrapperTask):
    """
    Just a wrapper to prepare all databases.
    """
    issue = luigi.Parameter(
        default=GBITask.DUMPTAG, description='tag to use as artificial "Dateissue" for dump')
    since = luigi.DateParameter(
        default=DUMP['date'], description='used in filename comparison')
    date = ClosestDateParameter(default=datetime.date.today())
    kind = luigi.Parameter(default='fulltext')

    def requires(self):
        prerequisite = GBIDatabaseList(kind=self.kind)
        luigi.build([prerequisite])
        with prerequisite.output().open() as handle:
            for row in handle.iter_tsv(cols=('db',)):
                yield GBIDatabase(issue=self.issue, since=self.since, date=self.date, db=row.db)

    def output(self):
        return self.input()


class GBIISSNStats(GBITask):
    """
    For each database list the percentage of records that have an ISSN.
    """
    issue = luigi.Parameter(
        default=GBITask.DUMPTAG, description='tag to use as artificial "Dateissue" for dump')
    since = luigi.DateParameter(
        default=DUMP['date'], description='used in filename comparison')
    date = ClosestDateParameter(default=datetime.date.today())
    kind = luigi.Parameter(default='fulltext')

    def requires(self):
        return GBIDatabaseISSNWrapper(issue=self.issue, since=self.since, date=self.date, kind=self.kind)

    def run(self):
        issn_pattern = re.compile(
            r'[0-9][0-9][0-9][0-9]-[0-9][0-9][0-9][0-9xX]')
        with self.output().open('w') as output:
            for target in self.input():
                stats = collections.defaultdict(int)
                issns = set()
                with target.open() as handle:
                    try:
                        for row in handle.iter_tsv():
                            if len(row) < 3:
                                stats['skipped'] += 1
                                continue
                            db, id, issn = row[0], row[1], row[2]

                            stats['db'] = db
                            stats['total'] += 1
                            if issn == 'NOT_AVAILABLE':
                                stats['noissn'] += 1
                            else:
                                stats['issn'] += 1
                                issns.add(issn)
                                if issn_pattern.search(issn):
                                    stats['valid'] += 1
                                else:
                                    stats['invalid'] += 1

                        stats['issncount'] = len(issns)
                        if stats['total'] > 0:
                            stats['withissn_percentage'] = '%0.2f' % (
                                100.0 / stats['total'] * stats['issn'])
                        else:
                            stats['withissn_percentage'] = 0
                        output.write_tsv(stats['db'], 'ok', stats['total'], stats['withissn_percentage'], stats['issncount'],
                                         stats['issn'], stats['noissn'], stats['valid'], stats['invalid'], stats['skipped'])
                    except Exception as err:
                        output.write_tsv(
                            stats['db'], 'err', 0, 0, 0, 0, 0, 0, 0, 0)
                        self.logger.debug(err)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)



# coding: utf-8
# pylint: disable=C0301,C0330,W0622,E1101

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""

CrossRef is an association of scholarly publishers
that develops shared infrastructure to support
more effective scholarly communications.

Our citation-linking network today covers over 68
million journal articles and other content items
(books chapters, data, theses, technical reports)
from thousands of scholarly and professional
publishers around the globe.

Configuration
-------------

[crossref]

doi-blacklist = /tmp/siskin-data/crossref/CrossrefDOIBlacklist/output.tsv

"""

import datetime
import json
import os
import tempfile
import time
import urllib

import luigi
import requests

import elasticsearch
from gluish.common import Executable
from gluish.format import TSV, Gzip
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import date_range, shellout
from siskin.benchmark import timed
from siskin.sources.amsl import AMSLFilterConfig, AMSLService
from siskin.task import DefaultTask
from siskin.utils import ElasticsearchMixin, URLCache


class CrossrefTask(DefaultTask):
    """
    Crossref related tasks. See: http://www.crossref.org/
    """
    TAG = 'crossref'

    def closest(self):
        """
        Update frequency.
        """
        return monthly(date=self.date)


class CrossrefHarvestChunkWithCursor(CrossrefTask):
    """
    Task should have the same output as `CrossrefHarvestChunk`, but
    implementation should use cursor (https://git.io/v1K27).
    """
    begin = luigi.DateParameter()
    end = luigi.DateParameter()
    filter = luigi.Parameter(default='deposit', description='index, deposit, update')

    rows = luigi.IntParameter(default=1000, significant=False)
    max_retries = luigi.IntParameter(default=10, significant=False, description='HTTP retries')
    attempts = luigi.IntParameter(default=3, significant=False, description='number of attempts to GET an URL that failed')
    sleep = luigi.IntParameter(default=1, significant=False, description='sleep between requests')

    def run(self):
        """
        > Using large offset values can result in extremely long response times. Offsets
        in the 100,000s and beyond will likely cause a timeout before the API is able
        to respond. An alternative to paging through very large result sets (like a
        corpus used for text and data mining) it to use the API's exposure of Solr's
        deep paging cursors. Any combination of query, filters and facets may be used
        with deep paging cursors. While rows may be specified along with cursor,
        offset and sample cannot be used. To use deep paging make a query as normal,
        but include the cursor parameter with a value of *
        """
        cache = URLCache(directory=os.path.join(tempfile.gettempdir(), '.urlcache'))
        adapter = requests.adapters.HTTPAdapter(max_retries=self.max_retries)
        cache.sess.mount('http://', adapter)

        filter = "from-{self.filter}-date:{self.begin},until-{self.filter}-date:{self.end}".format(self=self)
        rows, offset = self.rows, 0

        cursor = '*'

        with self.output().open('w') as output:
            while True:
                params = {
                    'rows': rows,
                    'filter': filter,
                    'cursor': cursor
                }

                url = 'http://api.crossref.org/works?%s' % (urllib.urlencode(params))

                for attempt in range(1, self.attempts):
                    if not cache.is_cached(url):
                        time.sleep(self.sleep)
                    body = cache.get(url)
                    try:
                        content = json.loads(body)
                    except ValueError as err:
                        if attempt == self.attempts - 1:
                            self.logger.debug('URL was %s', url)
                            self.logger.debug(err)
                            self.logger.debug(body[:100] + '...')
                            raise

                        cache_file = cache.get_cache_file(url)
                        if os.path.exists(cache_file):
                            self.logger.debug('trying to recover by removing cached entry at %s', cache_file)
                            os.remove(cache_file)
                    else:
                        break  # all fine

                count = len(content['message']['items'])
                self.logger.debug("%s: %s", url, count)
                if count == 0:
                    break

                output.write(body + '\n')
                offset += rows

                if not 'next-cursor' in content['message']:
                    raise RuntimeError('expected next-cursor in message, but missing')
                cursor = content['message']['next-cursor']

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))


class CrossrefHarvestChunk(CrossrefTask):
    """
    >>> DEPRECATED, switch to CrossrefHarvestChunkWithCursor soon, (https://git.io/v1K27).

    Harvest a slice of Crossref.

    API docs can be found under: http://api.crossref.org/

    The output file is line delimited JSON, just the concatenated responses.
    """
    begin = luigi.DateParameter()
    end = luigi.DateParameter()
    filter = luigi.Parameter(default='deposit', description='index, deposit, update')

    rows = luigi.IntParameter(default=200, significant=False)
    max_retries = luigi.IntParameter(default=10, significant=False, description='HTTP retries')
    attempts = luigi.IntParameter(default=3, significant=False, description='number of attempts to GET an URL that failed')
    sleep = luigi.IntParameter(default=1, significant=False, description='sleep between requests (secs)')

    @timed
    def run(self):
        """
        The API sometimes returns a 504 or other error. We therefore cache all HTTP requests
        locally with a simple URLCache and re-attempt a URL a couple of times.
        """
        cache = URLCache(directory=os.path.join(tempfile.gettempdir(), '.urlcache'))
        adapter = requests.adapters.HTTPAdapter(max_retries=self.max_retries)
        cache.sess.mount('http://', adapter)

        filter = "from-{self.filter}-date:{self.begin},until-{self.filter}-date:{self.end}".format(self=self)
        rows, offset = self.rows, 0

        with self.output().open('w') as output:
            while True:
                params = {"rows": rows, "offset": offset, "filter": filter}
                url = 'http://api.crossref.org/works?%s' % (urllib.urlencode(params))
                for attempt in range(1, self.attempts):
                    if not cache.is_cached(url):
                        time.sleep(self.sleep)
                    body = cache.get(url)
                    try:
                        content = json.loads(body)
                    except ValueError as err:
                        if attempt == self.attempts - 1:
                            self.logger.debug("URL was %s", url)
                            self.logger.debug(err)
                            self.logger.debug(body[:100])
                            raise
                        if os.path.exists(cache.get_cache_file(url)):
                            self.logger.debug("trying to recover by removing cached entry")
                            os.remove(cache.get_cache_file(url))
                    else:
                        break
                items = content["message"]["items"]
                self.logger.debug("%s: %s", url, len(items))
                if len(items) == 0:
                    break
                output.write(body + "\n")
                offset += rows

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))


class CrossrefHarvest(luigi.WrapperTask, CrossrefTask):
    """
    Harvest everything in incremental steps. Yield the targets sorted by date (ascending).
    """
    begin = luigi.DateParameter(default=datetime.date(2006, 1, 1))
    end = luigi.DateParameter(default=datetime.date.today())
    update = luigi.Parameter(default='months', description='days, weeks or months')

    def requires(self):
        if self.update not in ('days', 'weeks', 'months'):
            raise RuntimeError('update can only be: days, weeks or months')
        dates = [dt for dt in date_range(self.begin, self.end, 1, self.update)]
        tasks = [CrossrefHarvestChunkWithCursor(begin=dates[i], end=dates[i + 1])
                 for i in range(len(dates) - 1)]
        return sorted(tasks)

    def output(self):
        return self.input()


class CrossrefChunkItems(CrossrefTask):
    """
    Extract the message items, per chunk.
    """
    begin = luigi.DateParameter()
    end = luigi.DateParameter()
    filter = luigi.Parameter(default='deposit', description='index, deposit, update')

    def requires(self):
        return CrossrefHarvestChunkWithCursor(begin=self.begin, end=self.end, filter=self.filter)

    def run(self):
        output = shellout("jq -c -r '.message.items[]?' {input} | pigz -c > {output}", input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'), format=Gzip)


class CrossrefLineDOI(CrossrefTask):
    """
    Extract the line number and DOI for a given chunk. Works per chunk, so updates can be more incremental.

    TODO: Sort by filename seems correct for now, even in some edge cases, like
    DOI 10.1177/0959353515613812:

    13046   ... CrossrefChunkItems/begin-2016-01-01-end-2016-02-01-filter-deposit.ldj.gz 10.1177/0959353515613812
    3548977 ... CrossrefChunkItems/begin-2016-01-01-end-2016-02-01-filter-deposit.ldj.gz 10.1177/0959353515613812

    Both docs come in a single API harvest slice, but they differ slightly:
    https://gist.github.com/miku/c9e892343d8b0acb49cf

    Document in line #3548977 would be the most recent and the current
    implementation find it:

        $ taskcat CrossrefDOITable | LC_ALL=C grep -F "10.1177/0959353515613812"
        3548977 ... CrossrefChunkItems/begin-2016-01-01-end-2016-02-01-filter-deposit.ldj.gz 10.1177/0959353515613812

    So by using tac we should get the correct answer here, but it might probably
    be more robust, to use `timestamp` or some inherent attribute of the record
    for ordering.

    Q: DOI "10.1016/j.visres.2013.12.006" is in CrossrefChunkItems, but does not
    get extracted here. Why?

    A: 10.1016/j.visres.2013.12.006 is only an associated DOI, here a
    "Corrigendum to" the original article, which has been published in a
    subsequent volume. """

    begin = luigi.DateParameter()
    end = luigi.DateParameter()
    filter = luigi.Parameter(default='deposit', description='index, deposit, update')

    def requires(self):
        return CrossrefChunkItems(begin=self.begin, end=self.end, filter=self.filter)

    def run(self):
        output = shellout(r"""jq -r '.DOI?' <(unpigz -c {input}) | awk '{{print NR"\t{input}\t"$0 }}' | pigz -c > {output}""", input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='filelist.gz'))


class CrossrefLineDOIWrapper(CrossrefTask):
    """
    Run DOI and line extraction for each chunk.
    """
    begin = luigi.DateParameter(default=datetime.date(2006, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())
    update = luigi.Parameter(default='months', description='days, weeks or months')

    def requires(self):
        if self.update not in ('days', 'weeks', 'months'):
            raise RuntimeError('update can only be: days, weeks or months')
        dates = [dt for dt in date_range(self.begin, self.date, 1, self.update)]
        tasks = [CrossrefLineDOI(begin=dates[i - 1], end=dates[i]) for i in range(1, len(dates))]
        return sorted(tasks)

    def output(self):
        return self.input()


class CrossrefLineDOICombined(CrossrefTask):
    """
    Combine all extracted DOI and lines numbers from chunk files.
    Columns: filename lineno DOI.
    """
    begin = luigi.DateParameter(default=datetime.date(2006, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return CrossrefLineDOIWrapper(begin=self.begin, date=self.closest())

    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        for target in self.input():
            shellout("cat {input} >> {output}", input=target.path, output=stopover)
        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='filelist.gz'))


class CrossrefDOITable(CrossrefTask):
    """
    Create a table of current crossref DOIs along with their origin. Output
    contains each DOI only once, accociated with the most recent version.
    """
    begin = luigi.DateParameter(default=datetime.date(2006, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return CrossrefLineDOICombined(begin=self.begin, date=self.closest())

    def run(self):
        output = shellout("""
            TMPDIR={tmpdir} LC_ALL=C sort -k2,2 -S25% <(TMPDIR={tmpdir} unpigz -c {input}) |
            TMPDIR={tmpdir} tac |
            TMPDIR={tmpdir} LC_ALL=C sort -S25% -u -k3,3 |
            pigz -c > {output}""", tmpdir=self.config.get('core', 'tempdir'), input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='filelist.gz'), format=Gzip)


class CrossrefDOITableClean(CrossrefTask):
    """
    Clean the DOI table from DOIs, that are
    blacklisted (in crossref but leading nowhere).
    """
    begin = luigi.DateParameter(default=datetime.date(2006, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return CrossrefDOITable(begin=self.begin, date=self.date)

    def run(self):
        filepath = self.config.get('crossref', 'doi-blacklist', 'no-blacklist-available')

        # do not fail, if there is no blacklist configured
        if filepath == 'no-blacklist-available':
            self.logger.warning('no DOI blacklist configured, cf. #5706')
            self.input().copy(self.output().path)
            return

        # do not fail even, if file is configured but missing
        if not os.path.exists(filepath):
            self.logger.warn('crossref DOI blacklist configured, but file is missing: %s', filepath)
            self.input().copy(self.output().path)
            return

        blacklisted = set()

        with open(filepath) as handle:
            for line in (line.strip() for line in handle):
                blacklisted.add(line)

        with self.input().open() as handle:
            with self.output().open('w') as output:
                for line in (line.strip() for line in handle):
                    parts = line.split('\t')
                    if len(parts) < 3:
                        continue
                    if parts[2] in blacklisted:
                        continue
                    output.write(line + "\n")

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='filelist.gz'), format=Gzip)


class CrossrefSortedDOITable(CrossrefTask):
    """
    Sort the CrossrefDOITable by filename and line number.
    """
    begin = luigi.DateParameter(default=datetime.date(2006, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return CrossrefDOITableClean(begin=self.begin, date=self.closest())

    def run(self):
        output = shellout("""
            TMPDIR={tmpdir} LC_ALL=C sort -S50% -k2,2 -k1,1n <(TMPDIR={tmpdir} unpigz -c {input}) | cut -f 1-2 | pigz -c > {output}""",
                          tmpdir=self.config.get('core', 'tempdir'), input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='filelist.gz'), format=Gzip)


class CrossrefUniqItems(CrossrefTask):
    """
    What is left to do for calculating a Crossref
    snapshot: filter out the relevant lines.
    """
    begin = luigi.DateParameter(default=datetime.date(2006, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return CrossrefSortedDOITable(begin=self.begin, date=self.closest())

    def run(self):
        """
        We write line numbers to a separate file
        (L). We temporarily extract the chunk file
        as well (F), because pipes seem to fail
        randomly with many large files. Then we
        can run `filterline L F`.
        """
        _, stopover = tempfile.mkstemp(prefix='siskin-')

        previous_filename = None
        linenumbers = []

        with self.input().open() as handle:
            for line in handle:
                lineno, filename = line.strip().split('\t')
                if previous_filename and previous_filename != filename:
                    _, tmp = tempfile.mkstemp(prefix='siskin-')
                    with luigi.LocalTarget(tmp, format=TSV).open('w') as handle:
                        for lno in linenumbers:
                            handle.write_tsv(lno)

                    self.logger.info("filtering out %s lines", len(linenumbers))

                    input = shellout("unpigz -c {file} > {output}", file=previous_filename)
                    shellout("filterline {lines} {input} | pigz -c >> {output}",
                             filterline=self.assets('filterline'), lines=tmp, input=input, output=stopover)

                    try:
                        os.remove(tmp)
                        os.remove(input)
                    except OSError as err:
                        self.logger.warning(err)

                    linenumbers = []

                linenumbers.append(lineno)
                previous_filename = filename

        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'))


class CrossrefIntermediateSchema(CrossrefTask):
    """
    Convert to intermediate format via span.
    """
    begin = luigi.DateParameter(default=datetime.date(2006, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'span': Executable(name='span-import', message='http://git.io/vI8NV'),
            'file': CrossrefUniqItems(begin=self.begin, date=self.date)
        }

    @timed
    def run(self):
        output = shellout("span-import -i crossref <(unpigz -c {input}) | pigz -c > {output}",
                          input=self.input().get('file').path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'))


class CrossrefExport(CrossrefTask):
    """
    Tag with ISILs, then export to various formats.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    format = luigi.Parameter(default='solr5vu3', description='export format')

    def requires(self):
        return {
            'file': CrossrefIntermediateSchema(date=self.date),
            'config': AMSLFilterConfig(date=self.date),
        }

    def run(self):
        output = shellout("span-tag -c {config} <(unpigz -c {input}) | pigz -c > {output}",
                          config=self.input().get('config').path, input=self.input().get('file').path)
        output = shellout("span-export -o {format} <(unpigz -c {input}) | pigz -c > {output}", format=self.format, input=output)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        extensions = {
            'solr5vu3': 'ldj.gz',
            'formeta': 'form.gz',
        }
        return luigi.LocalTarget(path=self.path(ext=extensions.get(self.format, 'gz')))


class CrossrefCollections(CrossrefTask):
    """
    A collection of crossref collections, refs. #6985.
    """
    begin = luigi.DateParameter(default=datetime.date(2006, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'input': CrossrefIntermediateSchema(begin=self.begin, date=self.date),
                'jq': Executable(name='jq', message='https://github.com/stedolan/jq')}

    @timed
    def run(self):
        output = shellout("""jq -r '.["finc.mega_collection"]?' <(unpigz -c {input}) | LC_ALL=C sort -S35% -u > {output}""",
                          input=self.input().get('input').path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class CrossrefCollectionsDifference(CrossrefTask):
    """
    Refs. #7049. Check list of collections against AMSL Crossref collections and
    report difference.
    """
    begin = luigi.DateParameter(default=datetime.date(2006, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'crossref': CrossrefCollections(begin=self.begin, date=self.date),
            'amsl': AMSLService(date=self.date, name='outboundservices:discovery')
        }

    @timed
    def run(self):
        amsl = set()

        with self.input().get('amsl').open() as handle:
            items = json.load(handle)

        for item in items:
            if item['sourceID'] == '49':
                amsl.add(item['collectionLabel'].strip())

        with self.input().get('crossref').open() as handle:
            with self.output().open('w') as output:
                for row in handle.iter_tsv(cols=('name',)):
                    if row.name not in amsl:
                        output.write_tsv(row.name)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class CrossrefDOIList(CrossrefTask):
    """
    A list of Crossref DOIs.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'input': CrossrefIntermediateSchema(date=self.date),
                'jq': Executable(name='jq', message='https://github.com/stedolan/jq')}

    @timed
    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        # process substitution sometimes results in a broken pipe, so extract beforehand
        output = shellout("unpigz -c {input} > {output}", input=self.input().get('input').path)
        shellout("""jq -r '.doi?' {input} | grep -o "10.*" 2> /dev/null | LC_ALL=C sort -S50% > {output} """,
                 input=output, output=stopover)
        os.remove(output)
        output = shellout("""sort -S50% -u {input} > {output} """, input=stopover)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class CrossrefISSNList(CrossrefTask):
    """
    Just dump a list of all ISSN values. With dups and all.
    """
    begin = luigi.DateParameter(default=datetime.date(2006, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'input': CrossrefUniqItems(begin=self.begin, date=self.date),
                'jq': Executable(name='jq', message='https://github.com/stedolan/jq')}

    @timed
    def run(self):
        output = shellout("jq -r '.ISSN[]?' <(unpigz -c {input}) 2> /dev/null > {output}", input=self.input().get('input').path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class CrossrefUniqISSNList(CrossrefTask):
    """
    Just dump a list of all ISSN values. Sorted and uniq.
    """
    begin = luigi.DateParameter(default=datetime.date(2006, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return CrossrefISSNList(begin=self.begin, date=self.date)

    @timed
    def run(self):
        output = shellout("sort -u {input} > {output}", input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class CrossrefDOIAndISSNList(CrossrefTask):
    """
    A list of Crossref DOIs with their ISSNs.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'input': CrossrefIntermediateSchema(date=self.date),
                'jq': Executable(name='jq', message='https://github.com/stedolan/jq')}

    @timed
    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        temp = shellout("unpigz -c {input} > {output}", input=self.input().get('input').path)
        output = shellout("""jq -r '[.doi?, .["rft.issn"][]?, .["rft.eissn"][]?] | @csv' {input} | LC_ALL=C sort -S50% > {output} """,
                          input=temp, output=stopover)
        os.remove(temp)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='csv'))


class CrossrefIndex(CrossrefTask, ElasticsearchMixin):
    """
    Vanilla records into elasticsearch. All or
    nothing.
    """
    begin = luigi.DateParameter(default=datetime.date(2006, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())
    index = luigi.Parameter(default='crossref')

    def requires(self):
        return CrossrefUniqItems(begin=self.begin, date=self.date)

    @timed
    def run(self):
        es = elasticsearch.Elasticsearch()
        shellout("curl -XDELETE {host}:{port}/{index}", host=self.es_host, port=self.es_port, index=self.index)
        mapping = {
            'default': {
                'date_detection': False,
                '_id': {
                    'path': 'URL'
                },
            }
        }

        es.indices.create(index=self.index)
        es.indices.put_mapping(index=self.index, doc_type='default', body=mapping)
        shellout("esbulk -verbose -index {index} {input}", index=self.index, input=self.input().path)

        with self.output().open('w'):
            pass

    def output(self):
        return luigi.LocalTarget(path=self.path())


class CrossrefHarvestGeneric(CrossrefTask):
    """
    Basic harvest of members, funders, etc.
    """
    begin = luigi.DateParameter(default=datetime.date(2006, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())
    kind = luigi.Parameter(default='members')

    rows = luigi.IntParameter(default=1000, significant=False)
    max_retries = luigi.IntParameter(default=10, significant=False)

    @timed
    def run(self):
        cache = URLCache(directory=os.path.join(tempfile.gettempdir(), '.urlcache'))
        adapter = requests.adapters.HTTPAdapter(max_retries=self.max_retries)
        cache.sess.mount('http://', adapter)

        rows, offset = self.rows, 0

        with self.output().open('w') as output:
            while True:
                params = {"rows": rows, "offset": offset}
                url = 'http://api.crossref.org/%s?%s' % (self.kind, urllib.urlencode(params))
                for attempt in range(1, 3):
                    body = cache.get(url)
                    try:
                        content = json.loads(body)
                    except ValueError as err:
                        if attempt == 2:
                            self.logger.debug("URL was %s", url)
                            self.logger.debug(err)
                            self.logger.debug(body[:100])
                            raise
                        if os.path.exists(cache.get_cache_file(url)):
                            self.logger.debug("trying to recover by removing cached entry")
                            os.remove(cache.get_cache_file(url))
                    else:
                        break
                items = content["message"]["items"]
                self.logger.debug("%s: %s", url, len(items))
                if len(items) == 0:
                    break
                output.write(body)
                output.write("\n")
                offset += rows

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))


class CrossrefGenericItems(CrossrefTask):
    """
    Flatten and deduplicate. Stub.
    """
    begin = luigi.DateParameter(default=datetime.date(2006, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())
    kind = luigi.Parameter(default='members')

    def requires(self):
        return CrossrefHarvestGeneric(begin=self.begin, date=self.closest(), kind=self.kind)

    @timed
    def run(self):
        with self.output().open('w') as output:
            with self.input().open() as handle:
                for line in handle:
                    content = json.loads(line)
                    if not content.get("status") == "ok":
                        raise RuntimeError("invalid response status")
                    items = content["message"]["items"]
                    for item in items:
                        output.write(json.dumps(item))
                        output.write("\n")

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))


class CrossrefDOIHarvest(CrossrefTask):
    """
    Harvest DOI redirects from doi.org API. This is a long running task. It's
    probably best to run this tasks complete separate from the rest and let
    other processing pipelines use the result, as it is available.

    ----

    It is highly recommended that you put a static entry into /etc/hosts
    for doi.org while `hurrly` is running.

    As of 2015-07-31 doi.org resolves to six servers. Just choose one.

    $ nslookup doi.org

    """

    def requires(self):
        """
        If we have more DOI sources, we could add them as requirements here.
        """
        return {
            'input': CrossrefDOIList(date=datetime.date.today()),
            'hurrly': Executable(name='hurrly', message='http://github.com/miku/hurrly'),
            'pigz': Executable(name='pigz', message='http://zlib.net/pigz/')
        }

    def run(self):
        output = shellout("hurrly -w 64 < {input} | pigz > {output}", input=self.input().get('input').path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='tsv.gz'))


class CrossrefDOIBlacklist(CrossrefTask):
    """
    Create a blacklist of DOIs. Possible cases:

    1. A DOI redirects to http://www.crossref.org/deleted_DOI.html or
       most of these sites below crossref.org: https://gist.github.com/miku/6d754104c51fb553256d
    2. A DOI API lookup does not return a HTTP 200.

    The output of this task should be used as doi-blacklist in config.
    """

    def requires(self):
        return CrossrefDOIHarvest()

    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        shellout("""LC_ALL=C zgrep -E "http(s)?://.*.crossref.org" {input} >> {output}""",
                 input=self.input().path, output=stopover)
        shellout("""LC_ALL=C zgrep -v "^200" {input} >> {output}""",
                 input=self.input().path, output=stopover)
        output = shellout("sort -S50% -u {input} | cut -f4 | sed s@http://doi.org/api/handles/@@g > {output}", input=stopover)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103,C0301

# Copyright 2017 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Springer (testing).

Config
------

[springer]

intermediate-schema-file = /path/to/file # must be gzip compressed
"""

import luigi

from gluish.format import Gzip
from gluish.utils import shellout
from siskin.sources.amsl import AMSLFilterConfig
from siskin.task import DefaultTask


class SpringerTask(DefaultTask):
    TAG = 'springer'


class SpringerIntermediateSchema(SpringerTask, luigi.ExternalTask):
    """
    Provided.
    """

    def output(self):
        return luigi.LocalTarget(path=self.config.get('springer', 'intermediate-schema-file'))


class SpringerTagged(SpringerTask):
    """
    Tag records with ISIL.
    """

    def requires(self):
        return {
            'file': SpringerIntermediateSchema(),
            'config': AMSLFilterConfig(),
        }

    def run(self):
        output = shellout("span-tag -c {config} <(unpigz -c {input}) | pigz -c > {output}",
                          config=self.input().get('config').path,
                          input=self.input().get('file').path)

        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'), format=Gzip)


class SpringerExport(SpringerTask):
    """
    Solr importable.
    """
    format = luigi.Parameter(default='solr5vu3', description="solr5vu3, formeta")

    def requires(self):
        return SpringerTagged()

    def run(self):
        output = shellout("span-export <(unpigz -c {input}) | pigz -c > {output}", input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'))



# coding: utf-8
# pylint: disable=C0301,E1101,C0330

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Electronic Resource Management System Based on Linked Data Technologies.

http://amsl.technology

> Managing electronic resources has become a distinctive and important task for
libraries in recent years. The diversity of resources, changing licensing
policies and new business models of the publishers, consortial acquisition and
modern web scale discovery technologies have turned the market place of
scientific information into a complex and multidimensional construct.

Config:

[amsl]

uri-download-prefix = https://x.y.z/OntoWiki/files/get?setResource=
base = https://example.com
fid-issn-list = https://goo.gl/abcdef

"""

from __future__ import print_function

import collections
import datetime
import json
import operator
import tempfile
import zipfile

import luigi

from gluish.format import TSV
from gluish.utils import shellout
from siskin.task import DefaultTask
from siskin.utils import SetEncoder


class AMSLTask(DefaultTask):
    """
    Base class for AMSL related tasks.
    """
    TAG = 'amsl'


class AMSLService(AMSLTask):
    """
    Retrieve AMSL API response. Outbound: discovery, holdingsfiles,
    contentfiles, metadata_usage.

    Example output (discovery):

        [
            {
                "shardLabel": "SLUB-dbod",
                "sourceID": "64",
                "megaCollection": "Perinorm – Datenbank für Normen und technische Regeln",
                "productISIL": null,
                "externalLinkToContentFile": null,
                "contentFileLabel": null,
                "contentFileURI": null,
                "linkToContentFile": null,
                "ISIL": "DE-105",
                "evaluateHoldingsFileForLibrary": "no",
                "holdingsFileLabel": null,
                "holdingsFileURI": null,
                "linkToHoldingsFile": null
            },
            {
                "shardLabel": "SLUB-dbod",
                "sourceID": "64",
                "megaCollection": "Perinorm – Datenbank für Normen und technische Regeln",
                "productISIL": null,
                "externalLinkToContentFile": null,
                "contentFileLabel": null,
                "contentFileURI": null,
                "linkToContentFile": null,
                "ISIL": "DE-14",
                "evaluateHoldingsFileForLibrary": "no",
                "holdingsFileLabel": null,
                "holdingsFileURI": null,
                "linkToHoldingsFile": null
            },
        ...

    """
    date = luigi.DateParameter(default=datetime.date.today())
    name = luigi.Parameter(default='outboundservices:discovery',
                           description='discovery, holdingsfiles, contentfiles, metadata_usage')

    def run(self):
        parts = self.name.split(':')
        if not len(parts) == 2:
            raise RuntimeError('name must be of the form realm:name, e.g. outboundservices:discovery')
        realm, name = parts

        link = '%s/%s/list?do=%s' % (self.config.get('amsl', 'base').rstrip('/'), realm, name)
        output = shellout("""curl --fail "{link}" > {output} """, link=link)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(digest=True))


class AMSLCollectionsShardFilter(AMSLTask):
    """
    A per-shard list of collection entries. One record per line.

        {
          "evaluateHoldingsFileForLibrary": "no",
          "holdingsFileLabel": null,
          "megaCollection": "DOAJ Directory of Open Access Journals",
          "shardLabel": "UBL-ai",
          "contentFileURI": null,
          "sourceID": "28",
          "linkToHoldingsFile": null,
          "holdingsFileURI": null,
          "productISIL": null,
          "linkToContentFile": null,
          "contentFileLabel": null,
          "externalLinkToContentFile": null,
          "ISIL": "DE-14"
        }
        ....

    Shard distribution as of January 2017:

        $ taskcat AMSLService | jq -rc '.[] | .shardLabel' | sort | uniq -c | sort -nr
        53493 UBL-ai
         1121 UBL-main
          245 SLUB-dswarm
           19 SLUB-dbod

    """
    date = luigi.DateParameter(default=datetime.date.today())
    shard = luigi.Parameter(
        default='UBL-ai', description='only collect items for this shard')

    def requires(self):
        return AMSLService(date=self.date, name='outboundservices:discovery')

    def run(self):
        with self.input().open() as handle:
            doc = json.load(handle)

        with self.output().open('w') as output:
            for item in doc:
                if not item['shardLabel'] == self.shard:
                    continue
                output.write(json.dumps(item) + "\n")

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class AMSLCollectionsISILList(AMSLTask):
    """
    A per-shard list of ISILs for which AMSL has some information.

        DE-105
        DE-14
        DE-15
        DE-15-FID
        DE-1972
        DE-540
        ...
    """
    date = luigi.DateParameter(default=datetime.date.today())
    shard = luigi.Parameter(
        default='UBL-ai', description='only collect items for this shard')

    def requires(self):
        return AMSLService(date=self.date, name='outboundservices:discovery')

    def run(self):
        with self.input().open() as handle:
            doc = json.load(handle)

        isils = set()

        for item in doc:
            if not item['shardLabel'] == self.shard:
                continue
            isils.add(item['ISIL'])

        if len(isils) == 0:
            raise RuntimeError('no isils found: maybe mispelled shard name?')

        with self.output().open('w') as output:
            for isil in sorted(isils):
                output.write_tsv(isil)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class AMSLCollectionsISIL(AMSLTask):
    """
    Per ISIL list of collections.

        {
          "48": [
            "Genios (Recht)",
            "Genios (Sozialwissenschaften)",
            "Genios (Psychologie)",
            "Genios (Fachzeitschriften)",
            "Genios (Wirtschaftswissenschaften)"
          ],
          "49": [
            "Helminthological Society (CrossRef)",
            "International Association of Physical Chemists (IAPC) (CrossRef)",
            ...

    Examples (Jan 2017):

        $ taskcat AMSLCollectionsISIL --isil DE-Mit1 --shard SLUB-dbod | jq .
        {
            "64": [
                "Perinorm – Datenbank für Normen und technische Regeln"
            ]
        }

        $ taskcat AMSLCollectionsISIL --isil DE-14 --shard SLUB-dswarm | jq .
        {
            "83": [
                "SLUB/Mediathek"
            ],
            "69": [
                "Wiley ebooks"
            ],
            "105": [
                "Springer Journals"
            ],
            "94": [
                "Blackwell Publishing Journal Backfiles 1879-2005",
                "China Academic Journals (CAJ) Archiv",
                "Wiley InterScience Backfile Collections 1832-2005",
                "Torrossa / Periodici",
                "Cambridge Journals Digital Archive",
                "Elsevier Journal Backfiles on ScienceDirect",
                "Periodicals Archive Online",
                "Emerald Fulltext Archive Database",
                "Springer Online Journal Archives"
            ],
            "67": [
                "SLUB/Deutsche Fotothek"
            ]
        }

        $ taskcat AMSLCollectionsISIL --isil DE-15 --shard SLUB-dswarm | jq .
        {
            "68": [
                "OLC SSG Medien- / Kommunikationswissenschaft",
                "OLC SSG Film / Theater"
            ],
            "94": [
                "Blackwell Publishing Journal Backfiles 1879-2005",
                "China Academic Journals (CAJ) Archiv",
                "Wiley InterScience Backfile Collections 1832-2005",
                "Torrossa / Periodici",
                "Cambridge Journals Digital Archive",
                "Elsevier Journal Backfiles on ScienceDirect",
                "Periodicals Archive Online",
                "Emerald Fulltext Archive Database",
                "Springer Online Journal Archives"
            ],
            "105": [
                "Springer Journals"
            ]
        }
    """
    date = luigi.DateParameter(default=datetime.date.today())
    isil = luigi.Parameter(description='ISIL, case sensitive')
    shard = luigi.Parameter(default='UBL-ai', description='only collect items for this shard')

    def requires(self):
        return AMSLService(date=self.date, name='outboundservices:discovery')

    def run(self):
        with self.input().open() as handle:
            doc = json.load(handle)
        scmap = collections.defaultdict(set)
        for item in doc:
            if not item['shardLabel'] == self.shard:
                continue
            if not item['ISIL'] == self.isil:
                continue
        scmap[item['sourceID']].add(item['megaCollection'].strip())
        if not scmap:
            raise RuntimeError('no collections found for ISIL: %s' % self.isil)

        with self.output().open('w') as output:
            output.write(json.dumps(scmap, cls=SetEncoder) + "\n")

    def output(self):
        return luigi.LocalTarget(path=self.path())


class AMSLHoldingsFile(AMSLTask):
    """
    Access AMSL files/get?setResource= facilities.

    The output is probably zipped (will be decompressed on the fly).

    One ISIL can have multiple files (they will be combined).

    Output should be in standard KBART format, given the uploaded files in AMSL are KBART.
    """
    isil = luigi.Parameter(description='ISIL, case sensitive')
    date = luigi.Parameter(default=datetime.date.today())

    def requires(self):
        return AMSLService(date=self.date, name='outboundservices:holdingsfiles')

    def run(self):
        with self.input().open() as handle:
            holdings = json.load(handle)

        _, stopover = tempfile.mkstemp(prefix='siskin-')

        # The property which contains the URI of the holding file. Might
        # change.
        urikey = 'DokumentURI'

        for holding in holdings:
            if holding["ISIL"] == self.isil:

                if urikey not in holding:
                    raise RuntimeError('possible AMSL API change, expected: %s, available keys: %s' % (
                        urikey, holding.keys()))

                # refs. #7142
                if 'kbart' not in holding[urikey].lower():
                    self.logger.debug(
                        "skipping non-KBART holding URI: %s", holding[urikey])
                    continue

                link = "%s%s" % (self.config.get(
                    'amsl', 'uri-download-prefix'), holding[urikey])
                downloaded = shellout(
                    "curl --fail {link} > {output} ", link=link)
                try:
                    _ = zipfile.ZipFile(downloaded)
                    shellout("unzip -p {input} >> {output}",
                             input=downloaded, output=stopover)
                except zipfile.BadZipfile:
                    # at least the file is not a zip.
                    shellout("cat {input} >> {output}",
                             input=downloaded, output=stopover)

        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())


class AMSLOpenAccessISSNList(AMSLTask):
    """
    List of ISSN, which are open access.

    For now, extract these ISSN from a special holding file, called
    KBART_FREEJOURNALS via AMSL.

    Example (Jan 2017):

        $ taskcat AMSLOpenAccessISSNList | head -10
        0001-0944
        0001-1843
        0001-186X
        0001-1983
        0001-2114
        0001-2211
        0001-267X
        0001-3714
        0001-3757
        0001-3765
    """

    date = luigi.Parameter(default=datetime.date.today())

    def run(self):
        """
        Download, maybe unzip, grab column 2 and 3, keep ISSN, sort -u.
        """
        key = "http://amsl.technology/discovery/metadata-usage/Dokument/KBART_FREEJOURNALS"
        link = "%s%s" % (self.config.get('amsl', 'uri-download-prefix'), key)

        downloaded = shellout("curl --fail {link} > {output} ", link=link)
        _, stopover = tempfile.mkstemp(prefix='siskin-')

        try:
            _ = zipfile.ZipFile(downloaded)
            output = shellout("unzip -p {input} >> {output}", input=downloaded)
        except zipfile.BadZipfile:
            # at least the file is not a zip.
            output = shellout("cat {input} >> {output}", input=downloaded)

        shellout("cut -f 2 {input} | grep -oE '[0-9]{{4,4}}-[xX0-9]{{4,4}}' >> {output}", input=output, output=stopover)
        shellout("cut -f 3 {input} | grep -oE '[0-9]{{4,4}}-[xX0-9]{{4,4}}' >> {output}", input=output, output=stopover)
        output = shellout("sort -u {input} > {output}", input=stopover)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())


class AMSLFilterConfig(AMSLTask):
    """
    Next version of AMSL filter config.

    Cases:

    61084 Case 3 (ISIL, SID, Collection, Holding File)
      805 Case 2 (ISIL, SID, Collection, Product ISIL)
      380 Case 1 (ISIL, SID, Collection)
       38 Case 4 (ISIL, SID, Collection, External Content File)
       31 Case 6 (ISIL, SID, Collection, External Content File, Holding File)
        1 Case 7 (ISIL, SID, Collection, Internal Content File, Holding File)
        1 Case 5 (ISIL, SID, Collection, Internal Content File)

    Process:

                AMSL Discovery API
                      |
                      v
                AMSLFilterConfig
                      |
                      v
        $ span-tag -c config.json < input.is > output.is

    Notes:

    This task turns an AMSL API response into a filterconfig, which span-tag
    can understand.

    AMSL API might not specify everything we need to know, so this is the
    *only* place, where additional information can be added.

    Also, span-tag is fast, but not that fast, that we can iterate over a
    disjuction of 60K items for each of the 100M documents fast enough,
    which, if we could, would simplify the implementation of this task.

    The main speed improvement comes from using lists of collection names
    instead of having each collection processed separately - which is how it
    works conceptually: Each collection could use a separate KBART file (or use
    none at all).

    We ignore collection names, if (external) content files are used. These
    content files are usually there, because the data source contains
    collections, which cannot be determined by the datum itself.

    For SID 48 we test simple attachments via DB names parsed from KBART (span
    0.1.144 or later), refs. #10266.

    Performance data point: 22 ISIL each with between 1 and 26 alternatives for
    attachment, each alternative consisting of around three filters. Around 30
    holding or content files each with between 10 and 50000 entries referenced
    220 times in total: around 20k records/s.
    """

    date = luigi.Parameter(default=datetime.date.today())

    def requires(self):
        return AMSLService(date=self.date)

    def run(self):
        with self.input().open() as handle:
            doc = json.loads(handle.read())

        # Case: sid and collection.
        isilsidcollections = collections.defaultdict(
            lambda: collections.defaultdict(set))

        # Case: sid, collection, holding file.
        isilsidfilecollections = collections.defaultdict(
            lambda: collections.defaultdict(
                lambda: collections.defaultdict(set)))

        # Ready-made filters per ISIL:
        isilfilters = collections.defaultdict(list)

        for item in doc:
            if (all(operator.itemgetter('sourceID',
                                        'megaCollection',
                                        'ISIL')(item)) and
                    not any(operator.itemgetter('linkToHoldingsFile',
                                                'linkToContentFile',
                                                'externalLinkToContentFile',
                                                'productISIL')(item))):
                # Case 1 (ISIL, SID, Collection)
                #
                # SID COLL ISIL LTHF LTCF ELTCF PI
                # --------------------------------
                # X   X    X    -    -    -     -
                #
                # Group collections by ISIL and SID, combine into:
                # {"and": [{"source": [SID]}, {"collection": [...]}]}

                isilsidcollections[item['ISIL']][item['sourceID']].add(item['megaCollection'])

            elif (all(operator.itemgetter('sourceID',
                                          'megaCollection',
                                          'ISIL',
                                          'productISIL')(item)) and
                    not any(operator.itemgetter('linkToContentFile',
                                                'externalLinkToContentFile')(item))):
                # Case 2 (ISIL, SID, Collection, Product ISIL)
                #
                # SID COLL ISIL LTHF LTCF ELTCF PI
                # --------------------------------
                # X   X    X    -    -    -     X

                self.logger.debug("ignoring case with product isil for AI %s, %s, %s",
                                  item['ISIL'], item['sourceID'], item['productISIL'])

            elif (all(operator.itemgetter('sourceID',
                                          'megaCollection',
                                          'ISIL',
                                          'linkToHoldingsFile')(item)) and
                    not any(operator.itemgetter('linkToContentFile',
                                                'externalLinkToContentFile',
                                                'productISIL')(item))):
                # Case 3 (ISIL, SID, Collection, Holding File)
                #
                # SID COLL ISIL LTHF LTCF ELTCF PI
                # --------------------------------
                # X   X    X    X    -    -     -
                #
                # Group collections by ISIL, SID and holding file, combine into:
                # {"and": [{"source": [SID]}, {"collection": [...]}, {"holdings": {"urls": [...]}}]}

                # Exception: SID 48 should not use collection name, combine into:
                # {"and": [{"source": [SID]}, {"holdings": {"urls": [...]}}]}

                if item.get('evaluateHoldingsFileForLibrary') == "yes":
                    isilsidfilecollections[item['ISIL']][item['sourceID']][item['linkToHoldingsFile']].add(item['megaCollection'])
                else:
                    self.logger.warning("evaluateHoldingsFileForLibrary is not yes and still there is a link: skipping %s", item)

            elif (all(operator.itemgetter('sourceID',
                                          'megaCollection',
                                          'ISIL',
                                          'externalLinkToContentFile')(item)) and
                    not any(operator.itemgetter('linkToHoldingsFile',
                                                'linkToContentFile',
                                                'productISIL')(item))):
                # Case 4 (ISIL, SID, Collection, External Content File)
                #
                # SID COLL ISIL LTHF LTCF ELTCF PI
                # --------------------------------
                # X   X    X    -    -    X     -
                #
                # Create a single item: ISIL, SID, content file:
                # {"and": [{"source": [SID]}, {"holdings": {"urls": [...]}}]}
                #
                # Content files are used, when the raw data does not carry its collection.

                isilfilters[item["ISIL"]].append({
                    "and": [
                        {"source": [item["sourceID"]]},
                        {"holdings": {"urls": [item["externalLinkToContentFile"]]}},
                    ]})

            elif (all(operator.itemgetter('sourceID',
                                          'megaCollection',
                                          'ISIL',
                                          'linkToContentFile')(item)) and
                    not any(operator.itemgetter('linkToHoldingsFile',
                                                'externalLinkToContentFile',
                                                'productISIL')(item))):
                # Case 5 (ISIL, SID, Collection, Internal Content File)
                #
                # SID COLL ISIL LTHF LTCF ELTCF PI
                # --------------------------------
                # X   X    X    -    X    -     -
                #
                # Create a single item: ISIL, SID, content file:
                # {"and": [{"source": [SID]}, {"holdings": {"urls": [...]}}]}
                #
                # Content files are used, when the raw data does not carry its collection.

                isilfilters[item["ISIL"]].append({
                    "and": [
                        {"source": [item["sourceID"]]},
                        {"holdings": {"urls": [item["linkToContentFile"]]}},
                    ]})

            elif (all(operator.itemgetter('sourceID',
                                          'megaCollection',
                                          'ISIL',
                                          'linkToHoldingsFile',
                                          'externalLinkToContentFile')(item)) and
                    not any(operator.itemgetter('linkToContentFile',
                                                'productISIL')(item))):
                # Case 6 (ISIL, SID, Collection, External Content File, Holding File)
                #
                # SID COLL ISIL LTHF LTCF ELTCF PI
                # --------------------------------
                # X   X    X    X    -    X     -
                #
                # Create a single item: ISIL, SID, content file, holding file:
                # {"and": [{"source": [SID]}, {"holdings": {"urls": [...]}}, {"holdings": {"urls": [...]}}]}
                #
                # First, the content file restricts the set of records, then
                # another holding file can restrict the attachments, e.g. list of
                # ISSN or other.

                if item.get('evaluateHoldingsFileForLibrary') == "yes":
                    isilfilters[item["ISIL"]].append({
                        "and": [
                            {"source": [item["sourceID"]]},
                            {"holdings": {"urls": [item["externalLinkToContentFile"]]}},
                            {"holdings": {"urls": [item["linkToHoldingsFile"]]}},
                        ]})
                else:
                    self.logger.warning("evaluateHoldingsFileForLibrary is not yes and still there is a link: skipping %s", item)

            elif (all(operator.itemgetter('sourceID',
                                          'megaCollection',
                                          'ISIL',
                                          'linkToHoldingsFile',
                                          'linkToContentFile')(item)) and
                    not any(operator.itemgetter('externalLinkToContentFile',
                                                'productISIL')(item))):
                # Case 7 (ISIL, SID, Collection, Internal Content File, Holding File)
                #
                # SID COLL ISIL LTHF LTCF ELTCF PI
                # --------------------------------
                # X   X    X    X    X    -     -
                #
                # Create a single item: ISIL, SID, content file, holding file:
                # {"and": [{"source": [SID]}, {"holdings": {"urls": [...]}}, {"holdings": {"urls": [...]}}]}
                #
                # First, the content file restricts the set of records, then
                # another holding file can restrict the attachments, e.g. list of
                # ISSN or other.

                if item.get('evaluateHoldingsFileForLibrary') == "yes":
                    isilfilters[item["ISIL"]].append({
                        "and": [
                            {"source": [item["sourceID"]]},
                            {"holdings": {"urls": [item["linkToContentFile"]]}},
                            {"holdings": {"urls": [item["linkToHoldingsFile"]]}},
                        ]})
                else:
                    self.logger.warning("evaluateHoldingsFileForLibrary is not yes and still there is a link: skipping %s", item)
            else:
                # Bail out, if none of the above cases holds.
                raise RuntimeError("unhandled combination of sid, collection and other parameters: %s", item)

        # Second pass for some cases.
        for isil, blob in isilsidcollections.items():
            for sid, colls in blob.items():
                if sid == "48":
                    # isilfilters[isil].append({"source": [sid]})
                    self.logger.debug("""suppress single {"source": [48]} filter for %s""", isil)
                    continue
                isilfilters[isil].append({"and": [{"source": [sid]}, {"collection": colls}]})

        for isil, blob in isilsidfilecollections.items():
            for sid, spec in blob.items():
                for link, colls in spec.items():
                    if sid == "48":
                        isilfilters[isil].append({"and": [{"source": [sid]}, {"holdings": {"urls": [link]}}]})
                        continue
                    isilfilters[isil].append({"and": [{"source": [sid]}, {"collection": colls}, {"holdings": {"urls": [link]}}]})

        # Final assembly.
        filterconfig = collections.defaultdict(dict)
        for isil, filters in isilfilters.items():
            if len(filters) == 1:
                filterconfig[isil] = filters[0]
            else:
                filterconfig[isil] = {"or": filters}

        with self.output().open('w') as output:
            json.dump(filterconfig, output, cls=SetEncoder)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='json'))
